﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.IO;
using PINQ;

namespace Anomaly_Detection
{
    class Program
    {
        public class PINQAgentRecord : PINQAgent
        {
            public double tally;
            public override bool apply(double epsilon)
            {
                tally += epsilon;
                Console.WriteLine("\tepsilon\t" + tally);

                return base.apply(epsilon);

            }
        }

        #region Helpful IO routines
        public static IEnumerable<string> ReadFile(string filename)
        {
            var input = new StreamReader(filename);
            while (!input.EndOfStream)
                yield return input.ReadLine();

            input.Close();
        }

        public static void WriteFile(string filename, IEnumerable<string> stream)
        {
            var output = new StreamWriter(filename);
            foreach (var element in stream)
                output.WriteLine(element);

            output.Close();
        }
        #endregion

        #region matrix operations
        public class Entry
        {
            public double row;
            public double col;
            public double val;

            public Entry(double r, double c, double v) { row = r; col = c; val = v; }

            public override string ToString()
            {
                return row.ToString() + "\t" + col.ToString() + "\t" + val.ToString();
            }
        }

        // compacts a matrix by aggregating each coordinate entry.
        public static IQueryable<Entry> Condense(IQueryable<Entry> a)
        {
            return a.GroupBy(x => new { x.row, x.col }, x => x.val, (k, l) => new Entry(k.row, k.col, l.Sum()));
        }

        // multiplies two matrices together. best if one is pre-partitioned by join key
        public static IQueryable<Entry> Multiply(IQueryable<Entry> a, IQueryable<Entry> b)
        {
            return Condense(a.Join(b, x => x.col, y => y.row, (x, y) => new Entry(x.row, y.col, x.val * y.val)));
        }

        // adds two matrices. very easy.
        public static IQueryable<Entry> Add(IQueryable<Entry> a, IQueryable<Entry> b)
        {
            return Condense(a.Concat(b));
        }

        // takes the transpose of a matrix. 
        public static IQueryable<Entry> Transpose(IQueryable<Entry> a)
        {
            return a.Select(x => new Entry(x.col, x.row, x.val));
        }

        // computes the inverse Cholesky factor, to orthonormalize v
        public static IQueryable<Entry> Cholesky(IQueryable<Entry> v)
        {
            var vtv = Multiply(Transpose(v), v).ToArray();
            var keys = vtv.Select(x => x.col).Distinct().OrderBy(x => x).ToArray();

            var result = keys.Select(x => new Entry(x, x, 1.0)).ToArray();

            foreach (var key in keys)
            {
                var row = vtv.Where(x => x.row == key && x.col >= key).OrderBy(x => x.col).ToArray();
                var norm2 = row.Where(x => x.col == key).First().val;

                result = Multiply(result.AsQueryable(), keys.Select(x => new Entry(x, x, 1.0))
                                                            .Concat(row.Select(x => new Entry(x.row, x.col, -x.val / norm2)))
                                                            .Concat(new Entry[] { new Entry(key, key, 1.0 / Math.Sqrt(norm2)) }).AsQueryable()).ToArray();

                var update = row.Join(row, x => true, y => true, (x, y) => new Entry(x.col, y.col, -x.val * y.val / norm2));

                vtv = Add(vtv.AsQueryable(), update.AsQueryable()).ToArray();
            }

            return result.AsQueryable();
        }

        // useful for diagnostic stuff. 
        public static void PrintMatrix(IQueryable<Entry> a)
        {
            Console.WriteLine();
            foreach (var entry in a)
                Console.WriteLine(entry);
        }

        // each iteration of orthogonal iteration multiplies and then orthonormalizes. over and over ...
        public static IQueryable<Entry> OrthogonalIterationStep(IQueryable<Entry> a, IQueryable<Entry> v)
        {
            // first we multiply v through A then A transpose
            var atav = Multiply(Transpose(a), Multiply(a, v));

            // then we orthonormalize the result
            return Multiply(atav, Cholesky(atav));
        }

        #endregion

        public static IEnumerable<string> SyntheticData()
        {
            var rng = new Random();

            while (true)
            {
                var link = rng.Next() % 47;    // Any string identifier will work
                var time = rng.Next() % 392;    // Any string identifier will work

                var load = rng.Next() % 1500;   // some packet length (at most 1500)

                yield return link + "\t" + time + "\t" + load;
            }
        }

        static void Main(string[] args)
        {
            Console.WriteLine("Test harness for PCA-based anomaly detection");

            // comment out 
            WriteFile("input.txt", SyntheticData().Take(100000));

            var input = ReadFile("input.txt");

            // link names and time names are assumed public. we could use PrefixSearch.Discover if we needed to.
            var links = input.Select(x => x.Split('\t')[0]).Distinct().ToArray();
            var times = input.Select(x => x.Split('\t')[1]).Distinct().ToArray();

            var data = new PINQueryable<string>(input.ToArray().AsQueryable(), new PINQAgentRecord()).Select(x => x.Split('\t'));

            // each epsilon value is divided by 1500, to translate byte counts into packet counts
            foreach (var epsilon in new double[] { 0.1, 1.0, 10.0, 10000000.0 })
            {
                // we simply measure and record the covariance matrix. easy!
                var output = new StreamWriter("svd-output-" + epsilon + ".txt");

                // partition load by link
                var parts = data.Select(x => new { link = x[0], time = x[1], load = Convert.ToDouble(x[2]) })
                                .Partition(links, x => x.link);
                foreach (var link in links)
                {
                    // partition per-link load by time slice
                    var subparts = parts[link].Partition(times, x => x.time);
                    foreach (var time in times)
                        output.WriteLine(link + "\t" + time + "\t" + (subparts[time].Sum(epsilon, x => x.load / 1500.0)));
                }

                output.Close();

                // reading the covariance matrix back, we apply principle components analysis.
                // this part uses no privacy technology, only privately measured loads on links.
                var a = ReadFile("svd-output-" + epsilon + ".txt").Select(x => x.Split('\t'))
                                                                  .Select(x => new Entry(x[0].GetHashCode(), x[1].GetHashCode(), Convert.ToDouble(x[2]))).ToArray()
                                                                  .AsQueryable();

                a = Transpose(a);

                var dims = 4;               // number of svd dimensions
                var Random = new System.Random();

                var v = a.Select(x => x.col)
                         .Distinct()
                         .SelectMany(x => Enumerable.Range(0, dims)
                                                    .Select(y => new Entry(x, y, Random.NextDouble())))
                         .ToArray()
                         .AsQueryable();

                // run some large number of orthogonal iteration steps
                foreach (var iteration in Enumerable.Range(0, 10))
                    v = OrthogonalIterationStep(a, v).ToArray().AsQueryable();

                var norms = Condense(a.Select(x => new Entry(x.row, 1, x.val * x.val)));
                var diffs = Condense(a.Select(x => new Entry(x.row, 1, x.val * x.val))
                                      .Concat(Multiply(a, v).Select(x => new Entry(x.row, 1, -x.val * x.val))));

                var diff = new Dictionary<double, double>();
                foreach (var entry in diffs)
                    diff.Add(entry.row, entry.val);

                var output2 = new StreamWriter("anomaly-output-" + epsilon + ".txt");
                foreach (var time in times.Select((x, i) => new { x = x, i = i }))
                    output2.WriteLine(time.i + "\t" + Math.Sqrt(diff[time.x.GetHashCode()]));

                output2.Close();
            }
        }
    }
}
