﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.IO;
using PINQ;
using DPTraceAnalysis;

namespace Clustering_TTLs
{
    class Program
    {
        public class PINQAgentRecord : PINQAgent
        {
            public double tally;
            public override bool apply(double epsilon)
            {
                tally += epsilon;
                Console.WriteLine("\tepsilon\t" + tally);

                return base.apply(epsilon);

            }
        }

        #region Helpful IO routines
        public static IEnumerable<string> ReadFile(string filename)
        {
            var input = new StreamReader(filename);
            while (!input.EndOfStream)
                yield return input.ReadLine();

            input.Close();
        }

        public static void WriteFile(string filename, IEnumerable<string> stream)
        {
            var output = new StreamWriter(filename);
            foreach (var element in stream)
                output.WriteLine(element);

            output.Close();
        }
        #endregion

        #region k-means helper functions
        public static IEnumerable<double[]> RandomVectors(int dimension)
        {
            var RNG = new Random();
            while (true)
            {
                yield return Enumerable.Range(0, dimension)
                                       .Select(x => RNG.NextDouble() - 0.5)
                                       .ToArray();
            }
        }
        public static double Distance2(double[] a, double[] b)
        {
            return a.Select((x, i) => x - b[i]).Select(x => x * x).Sum();
        }
        public static double[] NearestCenter(double[] vector, IEnumerable<double[]> centers)
        {
            var bestcenter = centers.First();
            foreach (var center in centers)
                if (Distance2(center, vector) < Distance2(bestcenter, vector))
                    bestcenter = center;

            return bestcenter;
        }
        #endregion

        // performs one iteration of the k-means update rule, consuming epsilon units of differential privacy
        public static IEnumerable<double[]> kMeansStep(PINQueryable<double[]> vectors, IEnumerable<double[]> centers, double epsilon)
        {
            // partition the data set by associating each vector with its nearest center.
            var parts = vectors.Partition(centers.ToArray(), x => NearestCenter(x, centers));
            foreach (var center in centers)
            {
                var newcenter = center;
                foreach (var index in Enumerable.Range(0, center.Length))
                    newcenter[index] = parts[center].Average(epsilon / center.Length, x => x[index]);

                yield return newcenter;
            }
        }

        public static IEnumerable<string> SyntheticData()
        {
            var rng = new Random();

            while (true)
            {
                var TTL = rng.Next() % 64;
                var IP = String.Format("{0}.{1}.{2}.{3}", rng.Next() % 256, rng.Next() % 256, rng.Next() % 256, rng.Next() % 256);

                yield return IP + "\t" + TTL;
            }
        }

        static void Main(string[] args)
        {
            Console.WriteLine("Test harness for TTL clustering");

            // comment this out to stop regenerating random input
            foreach (var index in Enumerable.Range(0, 15))
                WriteFile("input-" + index + ".txt", SyntheticData().Take(1000));

            // each input file represents a monitor and is read in to a PINQueryable
            var datas = Enumerable.Range(0, 15)
                                  .Select(index => new PINQueryable<string>(ReadFile("input-" + index + ".txt").ToArray().AsQueryable(), index == 0 ? new PINQAgentRecord() : null));

            // turn the "name value" strings into <name, value> pairs.
            var monitors = datas.Select(data => data.Select(text => text.Split(' '))
                                                    .Where(x => x.Length >= 2)
                                                    .Select(x => new { name = x[0], value = (Convert.ToDouble(x[1]) - 32) / 32 })
                                                    .Materialize())
                                .ToArray();

            // initialize a common starting point for all experiments. k = 9.
            var starting = RandomVectors(monitors.Length).Take(9).ToArray();

            // various epsilon settings (0.1X usual, to account for 10x iterations)
            foreach (var epsilon in new double[] { 0.01, 0.1, 1.0, 10.0, 10000.0 })
            {
                // first compute (noisily) the average TTL for each coordinate
                var averages = monitors.Select(x => x.Average(epsilon / 2.0, y => y.value))
                                       .ToArray();

                // puts the honeypot ID on each value, inserts the average for missing values, converts to an array
                var vectors = monitors.Select((w, i) => w.Select(x => new { x.name, x.value, index = i }))
                                      .Aggregate((x, y) => x.Concat(y))
                                      .GroupBy(x => x.name)
                                      .Select(x => Enumerable.Range(0, monitors.Length)
                                                             .GroupJoin(x, y => y, y => y.index, (y1, y2) => new { index = y1, value = y2.Count() > 0 ? y2.Average(z => z.value) : averages[y1] })
                                                             .OrderBy(y => y.index)
                                                             .Select(y => y.value)
                                                             .ToArray())
                                      .Materialize();

                // we have a bunch of vectors in place now. do k-means!
                var centers = starting.Select(x => x.ToArray()).ToArray();
                foreach (var iteration in Enumerable.Range(0, 10))
                    centers = kMeansStep(vectors, centers, epsilon / 2.0).ToArray();
                
                // largest possible squared distance, as all coordinates are [-1, +1]
                var scale = monitors.Length * 4.0;
                
                // report how well we did (we ignore the privacy cost of the evaluation).
                Console.WriteLine(scale * vectors.Average(epsilon, x => Distance2(x, NearestCenter(x, centers)) / scale));
            }
        }
    }
}
