﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.IO;
using PINQ;
using DPTraceAnalysis;

namespace CumulativeDensityFunctions
{
    class Program
    {
        public class PINQAgentRecord : PINQAgent
        {
            public double tally;
            public override bool apply(double epsilon)
            {
                tally += epsilon;
                Console.WriteLine("\tepsilon\t" + tally);

                return base.apply(epsilon);

            }
        }

        #region Helpful IO routines
        public static IEnumerable<string> ReadFile(string filename)
        {
            var input = new StreamReader(filename);
            while (!input.EndOfStream)
                yield return input.ReadLine();

            input.Close();
        }

        public static void WriteFile(string filename, IEnumerable<string> stream)
        {
            var output = new StreamWriter(filename);
            foreach (var element in stream)
                output.WriteLine(element);

            output.Close();
        }
        #endregion

        public static string IPtoHex(string IPAddress)
        {
            var array = IPAddress.Split('.')
                                 .Select(x => Convert.ToInt32(x))
                                 .ToArray();

            return String.Format("{0:X2}{1:X2}{2:X2}{3:X2}", array[0], array[1], array[2], array[3]);
        }

        public static IEnumerable<string> SyntheticData()
        {
            var rng = new Random();

            while (true)
            { 
                var time = Convert.ToUInt64(rng.Next()) * 1000000;
                var srcIP = String.Format("{0}.{1}.{2}.{3}", rng.Next() % 256, rng.Next() % 256, rng.Next() % 256, rng.Next() % 256);;
                var dstIP = String.Format("{0}.{1}.{2}.{3}", rng.Next() % 256, rng.Next() % 256, rng.Next() % 256, rng.Next() % 256);;
                var srcPort = rng.Next() % 65536;
                var dstPort = rng.Next() % 65536;
                var syn = (rng.Next() % 2 == 0) ? "S" : "N";
                var ack = (rng.Next() % 2 == 0) ? "A" : "N";

                var seqn = rng.Next();
                var dlen = rng.Next() % 1500;
                var ackn = rng.Next();

                yield return time + "\t" + srcIP + "\t" + dstIP + "\t" + srcPort + "\t" + dstPort + "\t" + syn + "\t" + ack + "\t" + seqn + "\t" + dlen + "\t" + ackn;
            }
        }

        static void Main(string[] args)
        {
            Console.WriteLine("Test harness for CDF computation");
          
            // comment out to avoid reproducing synthetic data
            WriteFile("input.txt", SyntheticData().Take(10000));

            var data = new PINQueryable<string>(ReadFile("input.txt").ToArray().AsQueryable(), new PINQAgentRecord());
            var packets = data.Select(x => x.Split('\t'))
                              .Select(x => new
                              {
                                  time = Convert.ToUInt64(x[0]),
                                  srcIP = IPtoHex(x[1]),
                                  dstIP = IPtoHex(x[2]),

                                  srcport = x[3],
                                  dstport = x[4],

                                  syn = (x[5] == "S"),
                                  ack = (x[6] == "A"),

                                  seqn = Convert.ToUInt32(x[7]),
                                  dlen = Convert.ToUInt32(x[8]),
                                  ackn = Convert.ToUInt32(x[9]),
                              }).Materialize();

            // do a battery of tests for each epsilon value
            foreach (var epsilon in new double[] { 0.1, 1.0, 10.0, 1000000000.0 })
            {
                // packet length and destination port distributions are easy to read out of packets
                WriteFile("cdf-lengths-" + epsilon + ".txt", CDF.PINQCDF2(packets.Select(x => Convert.ToInt32(x.dlen)), 2048, epsilon).Select((x, i) => i + "\t" + x));
                WriteFile("cdf-ports-" + epsilon + ".txt", CDF.PINQCDF2(packets.Select(x => Convert.ToInt32(x.dstport)), 65536, epsilon).Select((x, i) => i + "\t" + x));
                
                // loss rate can be estamated by the number of retransmissions. here: distinct v. total sequence numbers seen on each flow.
                var retransmits = packets.GroupBy(x => new { x.srcIP, x.srcport, x.dstIP, x.dstport })
                                         .Select(g => 1.0 - g.Select(x => x.seqn).Distinct().Count() / (double)g.Count())
                                         .Select(x => Convert.ToInt32(1024 * x));

                WriteFile("cdf-lossrate-" + epsilon + ".txt", CDF.PINQCDF2(retransmits, 1024, epsilon / 2.0).Select((x, i) => (i / 1024.0) + "\t" + x));

                // next we match up syns and acks to measure RTTs                
                var syns = packets.Where(x => x.syn);
                var acks = packets.Where(x => x.ack);

                // match syn and ack packets based on src/dst and ack'ing the next sequence number
                var synacks = syns.Join(acks, x => new { x.srcIP, x.srcport, x.dstIP, x.dstport, seqn = x.seqn + 1 },
                                              y => new { srcIP = y.dstIP, srcport = y.dstport, dstIP = y.srcIP, dstport = y.srcport, seqn = y.ackn },
                                              (x, y) => new { x = x, y = y });

                // syn -> ack time difference
                var a = synacks.Where(pair => pair.x.syn && !pair.x.ack && pair.y.syn && pair.y.ack)
                               .Select(pair => Convert.ToInt32((pair.y.time - pair.x.time) / 10000));

                // ack -> syn time difference
                var b = synacks.Where(pair => pair.x.syn && pair.x.ack && !pair.y.syn && pair.y.ack)
                               .Select(pair => Convert.ToInt32((pair.y.time - pair.x.time) / 10000));

                WriteFile("cdf-synacks-" + epsilon + ".txt", CDF.PINQCDF2(a, 512, epsilon / 2.0).Select((x, i) => i + "\t" + x));
                WriteFile("cdf-acksyns-" + epsilon + ".txt", CDF.PINQCDF2(b, 512, epsilon / 2.0).Select((x, i) => i + "\t" + x));
            }
        }
    }
}
