﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using PINQ;
using System.IO;

namespace DPTraceAnalysis
{
    /// <summary>
    /// Contains several routines for computing the cumulative density of a collection of integer values.
    /// </summary>
    public class CDF
    {
        /// <summary>
        /// Approximates the cumulative density by counting the number of records less than each integer in the range.
        /// </summary>
        /// <param name="data">Source collection of integer values</param>
        /// <param name="maximum">The largest integer in the resulting cdf</param>
        /// <param name="epsilon">The amount of privacy to be consumed</param>
        /// <returns>A list of counts, where entry i is the number of records less than i</returns>
        public static IEnumerable<double> PINQCDF1(PINQueryable<int> data, int maximum, double epsilon)
        {
            foreach (var x in Enumerable.Range(0, maximum))
                yield return data.Where(record => record <= x).Count(epsilon / maximum);
        }

        /// <summary>
        /// Approximates the cumulative density by counting the number of records at each integer, and summing the results.
        /// </summary>
        /// <param name="data">Source collection of integer values</param>
        /// <param name="maximum">The largest integer in the resulting cdf</param>
        /// <param name="epsilon">The amount of privacy to be consumed</param>
        /// <returns>A list of counts, where entry i is the number of records less than i</returns>
        public static IEnumerable<double> PINQCDF2(PINQueryable<int> data, int maximum, double epsilon)
        {
            var tally = 0.0;
            var parts = data.Partition(Enumerable.Range(0, maximum).ToArray(), x => x);
            foreach (var x in Enumerable.Range(0, maximum))
                yield return (tally += parts[x].Count(epsilon));
        }

        /// <summary>
        /// Approximates the cumulative density through a recursive procedure estimating counts of regions at different resolutions.
        /// </summary>
        /// <param name="data">Source collection of integer values</param>
        /// <param name="maximum">The largest integer in the resulting cdf</param>
        /// <param name="epsilon">The amount of privacy to be consumed</param>
        /// <returns>A list of counts, where entry i is the number of records less than i</returns>
        public static IEnumerable<double> PINQCDF3(PINQueryable<int> data, int maximum, double epsilon)
        {
            if (maximum == 1)
                yield return data.Count(epsilon);
            else
            {
                var parts = data.Partition(new bool[] { true, false }, x => x < maximum / 2);
                foreach (var value in PINQCDF3(parts[true], maximum / 2, epsilon))
                    yield return value;

                var count = parts[true].Count(epsilon);
                parts[false] = parts[false].Select(x => x - maximum / 2);
                foreach (var value in PINQCDF3(parts[false], maximum / 2, epsilon))
                    yield return value + count;
            }
        }

        /// <summary>
        /// Converts any sequence of doubles to a non-decreasing sequence, minimizing the L2 error.
        /// </summary>
        /// <param name="input">Input sequence of doubles</param>
        /// <returns>Non-decreasing sequence of minimal L2 distance</returns>
        public static IEnumerable<double> PAVA(IEnumerable<double> input)
        {
            var stack = new Stack<double[]>();
            foreach (var entry in input)
            {
                var active = new double[] { entry, 1.0 };
                while (stack.Count() > 0 && active[0] < stack.Peek()[0])
                {
                    var temp = stack.Pop();
                    active[0] = (active[0] * active[1] + temp[0] * temp[1]) / (active[1] + temp[1]);
                    active[1] = active[1] + temp[1];
                }

                stack.Push(active);
            }

            return stack.SelectMany(x => Enumerable.Repeat(x[0], Convert.ToInt32(x[1]))).Reverse();
        }
    }

    /// <summary>
    /// A routine for finding common strings in a collection, from a specified alphabet.
    /// </summary>
    public class PrefixSearch
    {
        /// <summary>
        /// Enumerates strings found in the input collection, by incrementally growing prefixes with large support.
        /// </summary>
        /// <param name="data">Source collection of strings</param>
        /// <param name="keys">Alphabet over which the strings are drawn</param>
        /// <param name="confidence">Number of standard deviations for a count to register as non-zero</param>
        /// <param name="epsilon">Amount of privacy consumed in position</param>
        /// <returns>A list of strings with non-trivial support in the input</returns>
        public static IEnumerable<string> Discover(PINQueryable<string> data, char[] keys, double confidence, double epsilon)
        {
            var candidates = new string[] { "" };

            // each round extends the prefixes, until none are sufficiently supported
            for (var length = 0; candidates.Count() > 0; length++)
            {
                // only operate on viable strings
                var hashset = new HashSet<string>(candidates);
                var parts = data.Where(x => x.Length >= length && hashset.Contains(x.Substring(0, length)))
                                .Partition(new bool[] { true, false }, x => x.Length == length);

                // for strings that might be exact matches, count and output
                foreach (var part in parts[true].Partition(candidates, x => x))
                    if (part.Value.Count(epsilon) > confidence / epsilon)
                        yield return part.Key;

                // strings that will not be exact matches contribute to the extension of prefixes.
                candidates = parts[false].Partition(candidates, x => x.Substring(0, length))
                                         .Where(x => x.Value.Count(epsilon) > confidence / epsilon)
                                         .SelectMany(x => keys.Select(y => x.Key + y))
                                         .ToArray();
            }
        }
    }

    /// <summary>
    /// A routine for finding frequent item subsets in a collection of sets
    /// </summary>
    public class FrequentItemSets
    {
        /// <summary>
        /// Finds frequently occuring subsets in the input. Sets are all over strings, for simplicity.
        /// </summary>
        /// <param name="input">Source collection of sets of strings</param>
        /// <param name="items">List of base strings of interest</param>
        /// <param name="confidence">Number of standard deviations for a count to register as non-zero</param>
        /// <param name="epsilon">Amount of privacy consumed in each level</param>
        /// <returns>A list of all sets with non-trivial support, increasing in cardinality until none remain</returns>
        public static IEnumerable<string[]> FindSets(PINQueryable<string[]> input, IEnumerable<string> items, double confidence, double epsilon)
        {
            // a reference set of string arrays, conserving memory as well
            var candidates = new Dictionary<string, string>(items.Count());
            foreach (var item in items.Select(x => "\t" + x))
                candidates.Add(item, item);

            // replace each input array by an array of singleton arrays, from above
            var supporters = input.Select(x => x.Where(y => candidates.ContainsKey("\t" + y))  // only if we have it around
                                                .Select(y => candidates["\t" + y])             // get the unique representation
                                                .ToArray())
                                  .Materialize();

            // continue until we run out of candidates
            for (var i = 0; candidates.Count() > 0; i++)
            {
                // produce evidence of support for each of the candidates
                var parts = supporters.Where(x => x.Length > 0)
                                      .Select(x => x.First())
                                      .Partition(candidates.Values.ToArray(), x => x);

                // thin out candidates without support
                foreach (var part in parts)
                    if (part.Value.Count(epsilon) < confidence / epsilon)
                        candidates.Remove(part.Key);

                // Console.WriteLine(candidates.Count() + " candidates at level " + i);

                // return supported sets of this size
                foreach (var candidate in candidates.Keys)
                    yield return candidate.Split('\t');

                // the expansion of our current sent of candidates
                var newcandidates = new Dictionary<string, string>();
                foreach (var newcandidate in candidates.Values.Join(candidates.Values,
                                                                    x => x.Substring(0, x.LastIndexOf('\t')),
                                                                    y => y.Substring(0, y.LastIndexOf('\t')),
                                                                    (x, y) => x + y.Substring(y.LastIndexOf('\t'))))
                    if (newcandidate.Split('\t')[i + 1].GetHashCode() < newcandidate.Split('\t')[i + 2].GetHashCode())
                        newcandidates.Add(newcandidate, newcandidate);

                // update supporters by self-joining supported sets of each with themselves.
                supporters = supporters.Select(x => x.Where(y => candidates.ContainsKey(y))
                                                     .GroupBy(y => y.Substring(0, y.LastIndexOf('\t')), y => y.Substring(y.LastIndexOf('\t')))
                                                     .SelectMany(p => p.Join(p, y => true, z => true, (y, z) => p.Key + y + z))
                                                     .Where(y => newcandidates.ContainsKey(y))
                                                     .Select(y => newcandidates[y])
                                                     .Distinct()
                                                     .ToArray())
                                       .Materialize();

                // update the candidate set
                candidates = newcandidates;
            }
        }
    }
}
