﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.IO;
using PINQ;
using DPTraceAnalysis;

namespace Stepping_Stones
{
    class Program
    {
        public class PINQAgentRecord : PINQAgent
        {
            public double tally;
            public override bool apply(double epsilon)
            {
                tally += epsilon;
                Console.WriteLine("\tepsilon\t" + tally);

                return base.apply(epsilon);

            }
        }

        #region Helpful IO routines
        public static IEnumerable<string> ReadFile(string filename)
        {
            var input = new StreamReader(filename);
            while (!input.EndOfStream)
                yield return input.ReadLine();

            input.Close();
        }

        public static void WriteFile(string filename, IEnumerable<string> stream)
        {
            var output = new StreamWriter(filename);
            foreach (var element in stream)
                output.WriteLine(element);

            output.Close();
        }
        #endregion

        #region conversions between various representations
        public static string HexToIPPort(string hex)
        {
            var hexs = "0123456789ABCDEF";

            var ints = hex.Select(x => hexs.IndexOf(x)).ToArray();

            var result = "";

            result = result + (16 * ints[0] + ints[1]) + ".";
            result = result + (16 * ints[2] + ints[3]) + ".";
            result = result + (16 * ints[4] + ints[5]) + ".";
            result = result + (16 * ints[6] + ints[7]) + ":";

            result = result + (4096 * ints[8] + 256 * ints[9] + 16 * ints[10] + ints[11]);

            return result;
        }

        public static string PorttoHex(string port)
        {
            return String.Format("{0:X4}", Convert.ToInt32(port));
        }
        public static string IPtoHex(string IPAddress)
        {
            var array = IPAddress.Split('.')
                                 .Select(x => Convert.ToInt32(x))
                                 .ToArray();

            return String.Format("{0:X2}{1:X2}{2:X2}{3:X2}", array[0], array[1], array[2], array[3]);
        }

        public static string MesstoHex(string mess)
        {
            var pair = mess.Split('-');
            var hex1 = IPtoHex(pair[0].Substring(0, pair[0].IndexOf(':')));
            var hex2 = IPtoHex(pair[1].Substring(0, pair[1].IndexOf(':')));

            var port1 = PorttoHex(pair[0].Substring(pair[0].IndexOf(':') + 1));
            var port2 = PorttoHex(pair[1].Substring(pair[1].IndexOf(':') + 1));

            return hex1 + port1 + hex2 + port2;
        }
        #endregion

        public static IEnumerable<string> SyntheticData()
        {
            var rng = new Random();

            while (true)
            {
                var time = Convert.ToUInt64(rng.Next()) * 1000000;

                var srcIP = String.Format("{0}.{1}.{2}.{3}", rng.Next() % 256, rng.Next() % 256, rng.Next() % 256, rng.Next() % 256);
                var dstIP = String.Format("{0}.{1}.{2}.{3}", rng.Next() % 256, rng.Next() % 256, rng.Next() % 256, rng.Next() % 256);

                var srcPort = rng.Next() % 65536;
                var dstPort = rng.Next() % 65536;

                yield return time + "\t" + srcIP + "\t" + dstIP + "\t" + srcPort + "\t" + dstPort;
            }
        }

        static void Main(string[] args)
        {
            Console.WriteLine("Test harness for stepping stone analysis");

            // comment out to stop regenerating random input
            WriteFile("input.txt", SyntheticData().Take(10000).ToArray().AsQueryable());

            //var input = new PINQueryable<string>(ReadFile(args[0]).AsQueryable(), new PINQAgentRecord());
            var input = new PINQueryable<string>(ReadFile("input.txt").ToArray().AsQueryable(), new PINQAgentRecord());

            // parsing input, scouring payload for fingerprintable content.
            var trace = input.Where(x => !x.StartsWith("#"))
                             .Select(x => x.Split(' '))
                             .Where(x => x[5] == "TCP")
                             .Select(x => new { time = x[1], source = x[3], target = x[4], srcport = x[6], dstport = x[7] })
                             .Select(x => new { time = Convert.ToInt64(x.time), flow = IPtoHex(x.source) + PorttoHex(x.srcport) + IPtoHex(x.target) + PorttoHex(x.dstport) })
                             .Materialize();

            var window = 10000000 / 25;  // 40 ms
            var t_idle = 10000000 / 8;   // 1/8 s

            var confidence = 3.0;

            // three different experiments to do
            foreach (var epsilon in new double[] { 0.1, 1.0, 10.0 })
            {
                // characters appearing in hex strings
                var keys = new char[] { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };
                
                // we will need the names of common flows, to name the possible participants in stepping stones.
                // we didn't count this against the analysis' cost, assuming names exist, but we certainly could have.
                var flows = PrefixSearch.Discover(trace.Select(x => x.flow).Materialize(), keys, 10.0, epsilon / 24.0)
                                        .Where(x => x.Length == 24)
                                        .ToArray();

                Console.WriteLine("flows named: " + flows.Length);

                // computing upticks. broken into two parts.
                // half of the upticks occur in the second half of their time bucket.
                var half1 = trace.GroupBy(x => new { x.flow, time = x.time / (2 * t_idle) })
                                 .Where(x => x.Last().time > (x.Key.time * t_idle * 2) + t_idle)
                                 .Where(x => x.Count() == 1 || (x.Last().time > x.ElementAt(x.Count() - 1).time + t_idle))
                                 .Select(x => x.Last());

                // the other half occur in the first half of their time bucket.
                // add half a bucket width to all times, do as above, then subtract it back out.
                var half2 = trace.Select(x => new { time = x.time + t_idle, x.flow })
                                 .GroupBy(x => new { time = x.time / (2 * t_idle), x.flow })
                                 .Where(x => x.Last().time > (x.Key.time * t_idle * 2) + t_idle)
                                 .Where(x => x.Count() == 1 || (x.Last().time > x.ElementAt(x.Count() - 1).time + t_idle))
                                 .Select(x => x.Last())
                                 .Select(x => new { time = x.time - t_idle, x.flow });
                                 
                // putting both together gives us all of the upticks.
                var upticks = half1.Concat(half2).Materialize();

                Console.WriteLine("upticks materialized");

                #region actual sliding window logic. scary, and not frequent itemset compatible.
                /*
                var inputs1 = upticks.GroupBy(x => x.time / (2 * delta))
                                     .Select(g => g.Join(g, x => true, y => true, (x, y) => new { x = x, y = y })
                                                   .Where(pair => pair.x.time < pair.y.time)
                                                   .Where(pair => pair.x.time + delta > pair.y.time))
                                     .Select(l => l.Select(pair => new string[] { pair.x.flow, pair.y.flow }));

                var inputs2 = upticks.GroupBy(x => (x.time + delta) / (2 * delta))
                                     .Select(g => g.Join(g, x => true, y => true, (x, y) => new { x = x, y = y })
                                                   .Where(pair => pair.x.time < pair.y.time && pair.x.time + delta > pair.y.time))
                                     .Select(l => l.Select(pair => new string[] { pair.x.flow, pair.y.flow }));
                 
                var truth = inputs1.Concat(inputs2).Materialize();
                /**/
                #endregion

                // count up the upticks for each flow
                var parts = upticks.Partition(flows, x => x.flow);
                var counts = new Dictionary<string, double>();
                foreach (var flow in flows)
                    counts.Add(flow, parts[flow].Count(epsilon / 4.0));

                Console.WriteLine("activations counted for each flow");

                // restricting flows to similar counts makes FIM work better.
                flows = flows.Where(x => counts[x] > 1200 && counts[x] < 1400)
                             .ToArray();

                // now we get to GroupBy and find frequent itemsets
                var groups = upticks.GroupBy(x => x.time / window)
                                    .Select(x => x.Select(y => y.flow)
                                                  .Distinct()
                                                  .ToArray())
                                    .Where(x => x.Count() > 1);

                // the cost here is only for the first two iterations; subsequent ones are discarded
                var candidates = FrequentItemSets.FindSets(groups, flows, confidence, epsilon / 8.0)
                                                 .Where(x => x.Length == 3)
                                                 .ToArray();

                Console.WriteLine("candidates computed");

                // record how many times each flow occurs.
                // used to scale measurement noise, to limit privacy cost of measurement.
                var occurs = new Dictionary<string, int>();
                foreach (var group in candidates.SelectMany(x => x.Skip(1)).GroupBy(x => x))
                    occurs.Add(group.Key, group.Count());

                // count intersections of candidates.
                // also not counted against analysis cost.
                foreach (var candidate in candidates)
                {
                    var upticksa = parts[candidate[1]].Select(x => x.time / (window));
                    var upticksb = parts[candidate[2]].Select(x => x.time / (window));

                    var count = Convert.ToInt32(upticksa.Intersect(upticksb).Count(epsilon));

                    Console.WriteLine(candidate[1] + " -> " + candidate[2] + ":\t" + count + "\t+/- " + (1.0 / epsilon));
                }
            }
        }
    }
}
