﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.IO;
using PINQ;
using DPTraceAnalysis;


namespace Worm_Detection
{
    class Program
    {
        public class PINQAgentRecord : PINQAgent
        {
            public double tally;
            public override bool apply(double epsilon)
            {
                tally += epsilon;
                Console.WriteLine("\tepsilon\t" + tally);

                return base.apply(epsilon);

            }
        }

        #region Helpful IO routines
        public static IEnumerable<string> ReadFile(string filename)
        {
            var input = new StreamReader(filename);
            while (!input.EndOfStream)
                yield return input.ReadLine();

            input.Close();
        }

        public static void WriteFile(string filename, IEnumerable<string> stream)
        {
            var output = new StreamWriter(filename);
            foreach (var element in stream)
                output.WriteLine(element);

            output.Close();
        }
        #endregion

        public static string IPtoHex(string IPAddress)
        {
            var array = IPAddress.Split('.')
                                 .Select(x => Convert.ToInt32(x))
                                 .ToArray();

            return String.Format("{0:X2}{1:X2}{2:X2}{3:X2}", array[0], array[1], array[2], array[3]);
        }

        // returns a list of hashes of substrings, to fingerprint invariant content
        public static IEnumerable<string> substrings(string payload, string port)
        {
            var window = 40;
            foreach (var index in Enumerable.Range(0, Math.Max(payload.Length - window, 0)))
            {
                var width = Math.Min(window, payload.Length - index);
                if (payload.Substring(index, width).GetHashCode() % 64 == 0)
                    yield return string.Format("{0:X8}", (payload.Substring(index, width) + "\t" + port).GetHashCode());
            }
        }
        
        // choosing a [pseudo]random element from a list
        public static T PickRandom<T>(IEnumerable<T> options)
        {
            return options.ElementAt(Math.Abs(options.GetHashCode()) % options.Count());
        }

        // generates a bunch of well-formed lines of input data.
        public static IEnumerable<string> SyntheticData()
        {
            var rng = new Random();

            var wormpayload = Enumerable.Range(0, 32)
                                        .Select(x => String.Format("{0:X2}", rng.Next()))
                                        .Aggregate((x, y) => x + y);

            var wormport = rng.Next() % 65536;

            while (true)
            {
                var srcIP = String.Format("{0}.{1}.{2}.{3}", rng.Next() % 256, rng.Next() % 256, rng.Next() % 256, rng.Next() % 256);
                var dstIP = String.Format("{0}.{1}.{2}.{3}", rng.Next() % 256, rng.Next() % 256, rng.Next() % 256, rng.Next() % 256);

                if (rng.Next() % 2 == 0)
                    yield return srcIP + "\t" + dstIP + "\t" + wormport + "\t" + wormpayload;
                else
                    yield return srcIP + "\t" + dstIP + "\t" + String.Format("{0}", rng.Next() % 65536) + "\t" + Enumerable.Range(0, 32)
                                                                                                                           .Select(x => String.Format("{0:X2}", rng.Next()))
                                                                                                                           .Aggregate((x, y) => x + y);
            }
        }

        static void Main(string[] args)
        {
            Console.WriteLine("Test harness for EarlyBird worm fingerprinting");

            // comment this out to stop regenerating synthetic input
            WriteFile("input.txt", SyntheticData().Take(10000));

            //var input = new PINQueryable<string>(ReadFile(args[0]).AsQueryable(), new PINQAgentRecord());
            var input = new PINQueryable<string>(ReadFile("input.txt").ToArray().AsQueryable(), new PINQAgentRecord());

            // parsing input, scouring payload for fingerprintable content.
            var data = input.Select(x => x.Split('\t'))
                            .Select(x => new { source = x[0], target = x[1], port = x[2], payload = x[3] })
                            .Select(x => new { source = IPtoHex(x.source), target = IPtoHex(x.target), datas = substrings(x.payload, x.port)})
                            .Where(x => x.datas.Count() > 0)
                            .Select(x => x.source + "\t" + x.target + "\t" + x.datas.Aggregate((y, z) => y + "\t" + z));

            // for records with fingerprinted content, pick one print randomly to use.
            var pinq = data.Select(x => x.Split('\t'))
                           .Where(x => x.Length > 2)
                           .Select(x => x[0] + "\t" + x[1] + "\t" + PickRandom(x.Skip(2)))
                           .Materialize();

            // characters used in hex strings
            var keys = new char[] { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };

            // conduct several experiments, varying epsilon values.
            foreach (var epsilon in new double[] { 0.1, 1.0, 10.0 })
            {
                var output = new StreamWriter("worm-counts-" + epsilon + ".txt");

                var strings = pinq.Select(x => x.Split('\t')[2]);

                // look for common payloads in the data set using frequent string mining
                var names = PrefixSearch.Discover(strings, keys, 5.0, epsilon / 16.0)
                                        .Where(x => x.Length == 8)
                                        .ToArray();

                Console.WriteLine("Distinct names: {0}", names.Count());

                // wander through each name we have, counting dispersion
                var parts = pinq.Partition(names, x => x.Split('\t')[2]);
                foreach (var name in names)
                {
                    var sources = Convert.ToInt32(parts[name].Select(y => y.Split('\t')[0]).Distinct().Count(epsilon / 4.0));
                    var targets = Convert.ToInt32(parts[name].Select(y => y.Split('\t')[1]).Distinct().Count(epsilon / 4.0));

                    output.WriteLine(name + "\t" + sources + "\t" + targets);
                }

                output.Close();
            }
        }
    }
}
